# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import json
import os
import sys
from os.path import exists
from os.path import join
from os.path import normcase
from os.path import pathsep
import pkg_resources
import warnings

from calmjs import cli
from calmjs import dist
from calmjs.utils import pretty_logging
from calmjs.utils import finalize_env
from calmjs.utils import which
from calmjs.testing import mocks
from calmjs.testing.mocks import MockProvider
from calmjs.testing.utils import create_fake_bin
from calmjs.testing.utils import fake_error
from calmjs.testing.utils import mkdtemp
from calmjs.testing.utils import remember_cwd
from calmjs.testing.utils import stub_check_interactive
from calmjs.testing.utils import stub_item_attr_value
from calmjs.testing.utils import stub_base_which
from calmjs.testing.utils import stub_mod_call
from calmjs.testing.utils import stub_mod_check_output
from calmjs.testing.utils import stub_os_environ

which_node = which('node')
isatty = sys.stdin.isatty()


class CliGenerateMergeDictTestCase(unittest.TestCase):

    def test_merge(self):
        result = cli.generate_merge_dict(
            ['key'], {'key': {'foo': 1}}, {'baz': 1}, {'key': {'bar': 1}})
        self.assertEqual(result, {'key': {
            'foo': 1,
            'bar': 1,
        }})

    def test_merge_multi(self):
        result = cli.generate_merge_dict(
            ['key', 'mm'],
            {'key': {'foo': 1}},
            {'mm': {'snek': 'best'}},
            {'key': {'foo': 2}})
        self.assertEqual(result, {
            'key': {'foo': 2},
            'mm': {'snek': 'best'},
        })

    def test_merge_none_matched(self):
        result = cli.generate_merge_dict(
            ['none', 'match'], {'key': 'foo'}, {'bar': 1}, {'key': 'bar'})
        self.assertEqual(result, {})

    def test_using_actual_use_case(self):
        spec1 = {
            'dependencies': {
                'jquery': '~3.0.0',
                'underscore': '~1.8.0',
            },
            'devDependencies': {
                'sinon': '~1.17.0'
            },
            'name': 'foo',
        }

        spec2 = {
            'dependencies': {
                'jquery': '~1.11.0',
            },
            'devDependencies': {},
            'name': 'bar',
        }

        answer = {
            'dependencies': {
                'jquery': '~1.11.0',
                'underscore': '~1.8.0',
            },
            'devDependencies': {
                'sinon': '~1.17.0'
            },
        }

        result = cli.generate_merge_dict(
            ('dependencies', 'devDependencies'), spec1, spec2)

        # Naturally, the 'name' is missing and will need to be
        # reconciled separately... will figure this out later.
        self.assertEqual(result, answer)


class CliDriverTestCase(unittest.TestCase):
    """
    Base cli driver class test case.
    """

    def setUp(self):
        self.cwd = mkdtemp(self)
        remember_cwd(self)
        os.chdir(self.cwd)

    def test_get_bin_version_long(self):
        stub_mod_check_output(self, cli)
        stub_base_which(self)
        self.check_output_answer = b'Some app v.1.2.3.4. All rights reserved'
        results = cli.get_bin_version('some_app')
        self.assertEqual(results, (1, 2, 3, 4))

    def test_get_bin_version_longer(self):
        stub_mod_check_output(self, cli)
        stub_base_which(self)
        # tags are ignored for now.
        self.check_output_answer = b'version.11.200.345.4928-what'
        results = cli.get_bin_version('some_app')
        self.assertEqual(results, (11, 200, 345, 4928))

    def test_get_bin_version_short(self):
        stub_mod_check_output(self, cli)
        stub_base_which(self)
        self.check_output_answer = b'1'
        results = cli.get_bin_version('some_app')
        self.assertEqual(results, (1,))

    def test_get_bin_version_unexpected(self):
        stub_mod_check_output(self, cli)
        stub_base_which(self)
        self.check_output_answer = b'Nothing'
        with pretty_logging(stream=mocks.StringIO()) as err:
            results = cli.get_bin_version('some_app')
        self.assertIn(
            "encountered unexpected error while trying to find version of "
            "'some_app'", err.getvalue())
        self.assertIsNone(results)

    def test_get_bin_version_no_bin(self):
        stub_mod_check_output(self, cli, fake_error(OSError))
        stub_base_which(self)
        with pretty_logging(stream=mocks.StringIO()) as err:
            results = cli.get_bin_version('some_app')
        self.assertIn("failed to execute 'some_app'", err.getvalue())
        self.assertIsNone(results)

    def test_node_no_path(self):
        stub_os_environ(self)
        os.environ['PATH'] = ''
        with pretty_logging(stream=mocks.StringIO()) as err:
            self.assertIsNone(cli.get_node_version())
        self.assertIn("failed to execute 'node'", err.getvalue())

    def test_node_version_mocked(self):
        stub_mod_check_output(self, cli)
        stub_base_which(self)
        self.check_output_answer = b'v0.10.25'
        version = cli.get_node_version()
        self.assertEqual(version, (0, 10, 25))

    # live test, no stubbing
    @unittest.skipIf(which_node is None, 'Node.js not found.')
    def test_node_version_get(self):
        version = cli.get_node_version()
        self.assertIsNotNone(version)

    def test_node_run_no_path(self):
        stub_os_environ(self)
        os.environ['PATH'] = ''
        with self.assertRaises(OSError):
            cli.node('process.stdout.write("Hello World!");')

    # live test, no stubbing
    # some of these may take a long time on Windows for some reason.
    @unittest.skipIf(which_node is None, 'Node.js not found.')
    def test_node_run(self):
        stdout, stderr = cli.node('process.stdout.write("Hello World!");')
        self.assertEqual(stdout, 'Hello World!')
        stdout, stderr = cli.node('window')
        self.assertIn('window is not defined', stderr)

    # live test, no stubbing
    @unittest.skipIf(cli.get_node_version() is None, 'Node.js not found.')
    def test_node_run_bytes(self):
        stdout, stderr = cli.node(b'process.stdout.write("Hello World!");')
        self.assertEqual(stdout, b'Hello World!')

    # Note that for the following tests with the mock 'mgr' binary, both
    # call calmjs.cli and calmjs.base.which are stubbed.  This is due to
    # how the _exec method uses the which function to locate the binary,
    # and as mgr is not there it will fail.  The alternative is to keep
    # creating a dummy binary, but the tests for those are already done
    # so we can skip that for with mocks for the following tests.

    def test_helper_attr(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        with self.assertRaises(AttributeError) as e:
            driver.no_such_attr_here
        self.assertIn('no_such_attr_here', str(e.exception))
        self.assertIsNot(driver.mgr_init, None)
        self.assertIsNot(driver.get_mgr_version, None)
        self.assertTrue(driver.mgr_install(['calmjs']))
        self.assertEqual(self.call_args[0], (['mgr', 'install'],))

    def test_install_failure(self):
        stub_mod_call(self, cli, fake_error(IOError))
        stub_base_which(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        with pretty_logging(stream=mocks.StringIO()) as stderr:
            with self.assertRaises(IOError):
                driver.mgr_install(['calmjs'])
        val = stderr.getvalue()
        self.assertIn("invocation of the 'mgr' binary failed", val)

    def test_install_arguments(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'], args=('--pedantic',))
        self.assertEqual(
            self.call_args[0], (['mgr', 'install', '--pedantic'],))

    def test_alternative_install_cmd(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', install_cmd='sync')
        driver.pkg_manager_install(['calmjs'])
        self.assertEqual(self.call_args[0], (['mgr', 'sync'],))

        # Naturally, the short hand call will be changed.
        # note that args is NOT the package_name, and thus this just
        # means that the installation may not operate as expected off
        # the package.
        driver.mgr_sync(['calmjs'], args=('all',))
        self.assertEqual(self.call_args[0], (['mgr', 'sync', 'all'],))

    def test_install_other_environ(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        stub_os_environ(self)
        # pop out NODE_PATH if available
        os.environ.pop('NODE_PATH', '')
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'], env={
                'MGR_ENV': 'production'})
        self.assertEqual(self.call_args, ((['mgr', 'install'],), {
            'env': finalize_env({'MGR_ENV': 'production'}),
        }))

    def test_set_node_path(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        node_path = mkdtemp(self)
        driver = cli.PackageManagerDriver(
            node_path=node_path, pkg_manager_bin='mgr')

        # ensure env is passed into the call.
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'])
        self.assertEqual(self.call_args, ((['mgr', 'install'],), {
            'env': finalize_env({'NODE_PATH': node_path}),
        }))

        # will be overridden by instance settings.
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'], env={
                'PATH': '.',
                'MGR_ENV': 'dev',
                'NODE_PATH': '/tmp/somewhere/else/node_mods',
            })
        self.assertEqual(self.call_args, ((['mgr', 'install'],), {
            'env': finalize_env(
                {'NODE_PATH': node_path, 'MGR_ENV': 'dev', 'PATH': '.'}),
        }))

    def test_predefined_path(self):
        # ensure that the various paths are passed to env or cwd.
        stub_mod_call(self, cli)
        stub_base_which(self)
        somepath = mkdtemp(self)
        cwd = self.cwd
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', env_path=somepath, working_dir=cwd)
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'])
        args, kwargs = self.call_args
        self.assertEqual(kwargs['env']['PATH'].split(pathsep)[0], somepath)
        self.assertEqual(kwargs['cwd'], cwd)

    def test_env_path_not_exist(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        bad_path = '/no/such/path/for/sure/at/here'
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', env_path=bad_path)
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'])
        args, kwargs = self.call_args
        self.assertNotEqual(kwargs['env']['PATH'].split(pathsep)[0], bad_path)

    def test_paths_unset(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'])
        args, kwargs = self.call_args
        self.assertNotIn('PATH', kwargs)
        self.assertNotIn('cwd', kwargs)

    def test_working_dir_set(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        some_cwd = self.cwd
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', working_dir=some_cwd)
        with pretty_logging(stream=mocks.StringIO()):
            driver.pkg_manager_install(['calmjs'])
        args, kwargs = self.call_args
        self.assertNotIn('PATH', kwargs)
        self.assertEqual(kwargs['cwd'], some_cwd)

    def test_set_binary_no_package(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='bower')
        with pretty_logging(stream=mocks.StringIO()) as fd:
            driver.pkg_manager_install()
            self.assertIn(
                "no package name supplied, "
                "not continuing with 'bower install'", fd.getvalue())
        self.assertIsNone(self.call_args)

    def test_set_binary_with_package(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='bower')
        # this will call ``bower install`` instead.
        driver.pkg_manager_install(['calmjs'])
        self.assertEqual(self.call_args[0], (['bower', 'install'],))

    def test_which_is_none(self):
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        self.assertIsNone(driver.which())
        driver.env_path = mkdtemp(self)
        self.assertIsNone(driver.which())

    def create_fake_mgr_bin(self, root):
        return create_fake_bin(root, 'mgr')

    def test_which_is_set(self):
        stub_os_environ(self)
        tmpdir = mkdtemp(self)
        mgr_bin = self.create_fake_mgr_bin(tmpdir)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        driver.env_path = tmpdir
        self.assertEqual(normcase(driver.which()), normcase(mgr_bin))

        driver.env_path = None
        self.assertIsNone(driver.which())

    def test_which_is_set_env_path(self):
        stub_os_environ(self)
        tmpdir = mkdtemp(self)
        mgr_bin = self.create_fake_mgr_bin(tmpdir)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        # With both env_path attr and environ PATH set
        os.environ['PATH'] = driver.env_path = tmpdir
        self.assertEqual(normcase(driver.which()), normcase(mgr_bin))

        # the autodetection should still work through ENV_PATH
        driver.env_path = None
        self.assertEqual(normcase(driver.which()), normcase(mgr_bin))

    def test_set_env_path_with_node_modules_fail(self):
        stub_os_environ(self)
        tmpdir = mkdtemp(self)
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', working_dir=tmpdir)
        self.assertFalse(driver._set_env_path_with_node_modules())
        self.assertIsNone(driver.env_path)
        self.assertIsNone(driver.which_with_node_modules())

    def fake_mgr_bin(self):
        tmpdir = mkdtemp(self)
        # fake an executable in node_modules
        bin_dir = join(tmpdir, 'node_modules', '.bin')
        os.makedirs(bin_dir)
        fake_bin = self.create_fake_mgr_bin(bin_dir)
        return tmpdir, bin_dir, fake_bin

    def test_set_env_path_with_node_modules_success(self):
        tmpdir, bin_dir, mgr_bin = self.fake_mgr_bin()
        # constructor with an explicit working directory.
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', working_dir=tmpdir)
        self.assertIsNone(driver.env_path)
        # the which_with_node_modules should work immediately
        self.assertEqual(
            normcase(driver.which_with_node_modules()), normcase(mgr_bin))
        self.assertTrue(driver._set_env_path_with_node_modules())
        self.assertEqual(driver.env_path, bin_dir)
        # should still result in the same thing.
        self.assertTrue(driver._set_env_path_with_node_modules())
        self.assertEqual(driver.env_path, bin_dir)

    def test_set_env_path_with_node_path_success(self):
        tmpdir, bin_dir, mgr_bin = self.fake_mgr_bin()
        other_dir = mkdtemp(self)
        # default constructor
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', working_dir=other_dir)
        # the which_with_node_modules will not work immeidately in this
        # case
        self.assertIsNone(driver.which_with_node_modules())
        self.assertIsNone(driver.env_path)
        # using NODE_PATH set to a valid node_modules
        driver.node_path = join(tmpdir, 'node_modules')
        # this should work now.
        self.assertEqual(
            normcase(driver.which_with_node_modules()), normcase(mgr_bin))
        self.assertTrue(driver._set_env_path_with_node_modules())
        self.assertEqual(driver.env_path, bin_dir)
        # should still result in the same thing.
        self.assertTrue(driver._set_env_path_with_node_modules())
        self.assertEqual(driver.env_path, bin_dir)

    def test_set_env_path_with_node_path_with_environ(self):
        stub_os_environ(self)
        tmpdir, bin_dir, mgr_bin = self.fake_mgr_bin()
        # define a NODE_PATH set to a valid node_modules
        os.environ['NODE_PATH'] = join(tmpdir, 'node_modules')
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        self.assertTrue(driver._set_env_path_with_node_modules())
        self.assertEqual(driver.env_path, bin_dir)

    def test_set_env_path_with_node_path_multiple_with_environ(self):
        tmp = mkdtemp(self)
        tmp1, bin_dir1, _ = self.fake_mgr_bin()
        tmp2, bin_dir2, _ = self.fake_mgr_bin()
        node_path = pathsep.join(
            join(d, 'node_modules') for d in (tmp, tmp1, tmp2))
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', node_path=node_path)
        self.assertTrue(driver._set_env_path_with_node_modules())
        # First one.  Whether the node modules loads correctly, that's
        # up to the nodejs circus.
        self.assertEqual(driver.env_path, bin_dir1)

        # ensure the kws generated correctly.
        env = driver._gen_call_kws()['env']
        self.assertEqual(env['NODE_PATH'], node_path)
        self.assertEqual(env['PATH'].split(pathsep)[0], bin_dir1)

    def test_driver_run_failure(self):
        # testing for success may actually end up being extremely
        # annoying, so we are going to avoid that and let the integrated
        # subclasses deal with it.
        stub_os_environ(self)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        os.environ['PATH'] = ''
        with self.assertRaises(OSError):
            driver.run()

    # Helpers for getting a module level default instance up

    def test_driver_create_failure(self):
        with self.assertRaises(TypeError):
            # can't create the parent one as it is not subclassed like
            # the following
            cli.PackageManagerDriver.create()

    def test_driver_create(self):
        class Driver(cli.PackageManagerDriver):
            def __init__(self, **kw):
                kw['pkg_manager_bin'] = 'mgr'
                super(Driver, self).__init__(**kw)

        inst = Driver.create()
        self.assertTrue(isinstance(inst, Driver))

    def test_create_for_module_vars_and_warning(self):
        stub_os_environ(self)
        tmpdir = mkdtemp(self)
        values = {}

        class MgrDriver(cli.PackageManagerDriver):
            def __init__(self, *a, **kw):
                kw['pkg_manager_bin'] = 'mgr'
                kw['working_dir'] = tmpdir
                super(MgrDriver, self).__init__(*a, **kw)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            driver = MgrDriver.create_for_module_vars(values)
            self.assertTrue(issubclass(w[-1].category, RuntimeWarning))
            self.assertIn(
                "Unable to locate the 'mgr' binary or runtime",
                str(w[-1].message))

        self.assertTrue(isinstance(driver, MgrDriver))
        # Normally, these will be global names.
        self.assertIn('mgr_install', values)
        self.assertIn('mgr_init', values)
        self.assertIn('get_mgr_version', values)

    # Should really put more tests of these kind in here, but the more
    # concrete implementations have done so.  This weird version here
    # is mostly just for laughs.

    def setup_requirements_json(self):
        # what kind of bizzaro world do the following users live in?
        requirements = {"require": {"setuptools": "25.1.6"}}
        mock_provider = MockProvider({
            'requirements.json': json.dumps(requirements),
        })
        # seriously lolwat?
        mock_dist = pkg_resources.Distribution(
            metadata=mock_provider, project_name='calmpy.pip', version='0.0.0')
        working_set = pkg_resources.WorkingSet()
        working_set.add(mock_dist)
        stub_item_attr_value(self, dist, 'default_working_set', working_set)
        return working_set

    def test_pkg_manager_view(self):
        self.setup_requirements_json()
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
        )
        result = driver.pkg_manager_view('calmpy.pip')
        self.assertEqual(result, {
            "require": {"setuptools": "25.1.6"},
            "name": "calmpy.pip",
        })
        result2 = driver.mgr_view('calmpy.pip')
        self.assertEqual(result, result2)

    def test_pkg_manager_init(self):
        # we still need a temporary directory, but the difference is
        # that whether the instance contains it or not.
        self.setup_requirements_json()
        cwd = self.cwd

        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
        )
        driver.pkg_manager_init('calmpy.pip')

        target = join(cwd, 'requirements.json')
        self.assertTrue(exists(target))
        with open(target) as fd:
            result = json.load(fd)
        self.assertEqual(result, {
            "require": {"setuptools": "25.1.6"},
            "name": "calmpy.pip",
        })

    def test_pkg_manager_init_working_dir(self):
        self.setup_requirements_json()
        original = mkdtemp(self)
        os.chdir(original)
        cwd = mkdtemp(self)
        target = join(cwd, 'requirements.json')

        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
            working_dir=cwd,
        )
        driver.pkg_manager_init('calmpy.pip')

        self.assertFalse(exists(join(original, 'requirements.json')))
        self.assertTrue(exists(target))

        with open(target) as fd:
            result = json.load(fd)
        self.assertEqual(result, {
            "require": {"setuptools": "25.1.6"},
            "name": "calmpy.pip",
        })

    def test_pkg_manager_init_exists_and_overwrite(self):
        self.setup_requirements_json()
        cwd = mkdtemp(self)
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
            working_dir=cwd,
        )
        target = join(cwd, 'requirements.json')
        with open(target, 'w') as fd:
            result = json.dump({"require": {}}, fd)

        with pretty_logging(stream=mocks.StringIO()) as err:
            driver.pkg_manager_init('calmpy.pip', overwrite=False)

        self.assertIn('not overwriting existing ', err.getvalue())
        self.assertIn('requirements.json', err.getvalue())
        with open(target) as fd:
            result = json.load(fd)
        self.assertNotEqual(result, {"require": {"setuptools": "25.1.6"}})

        stub_mod_call(self, cli)
        with pretty_logging(stream=mocks.StringIO()) as err:
            # ensure the return value is False
            self.assertFalse(
                driver.pkg_manager_install('calmpy.pip', overwrite=False))

        driver.pkg_manager_init('calmpy.pip', overwrite=True)
        with open(target) as fd:
            result = json.load(fd)
        self.assertEqual(result, {
            "require": {"setuptools": "25.1.6"},
            "name": "calmpy.pip",
        })

    def test_pkg_manager_init_merge(self):
        self.setup_requirements_json()
        cwd = mkdtemp(self)
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
            working_dir=cwd,
        )
        target = join(cwd, 'requirements.json')
        with open(target, 'w') as fd:
            result = json.dump({"require": {"calmpy": "1.0.0"}}, fd)

        driver.pkg_manager_init('calmpy.pip', merge=True, overwrite=True)
        self.assertNotEqual(result, {
            "require": {
                "calmpy": "1.0.0",
                "setuptools": "25.1.6",
            },
            "name": "calmpy.pip",
        })

        stub_mod_call(self, cli)
        stub_base_which(self)
        with pretty_logging(stream=mocks.StringIO()):
            # ensure the return value is True, assuming successful
            self.assertTrue(
                driver.pkg_manager_install('calmpy.pip', overwrite=True))

    def test_pkg_manager_view_requires(self):
        working_set = self.setup_requirements_json()
        working_set.add(pkg_resources.Distribution(
            metadata=MockProvider({
                'requires.txt': 'calmpy.pip',
            }),
            project_name='site',
            version='0.0.0',
        ))
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
        )
        result = driver.pkg_manager_view('site')
        self.assertEqual(result, {
            "require": {"setuptools": "25.1.6"},
            "name": "site",
        })
        # try explicit
        result = driver.pkg_manager_view('site', explicit=True)
        self.assertEqual(result, {
            "require": {},
            "name": "site",
        })

    def test_pkg_manager_view_extras_requires(self):
        working_set = self.setup_requirements_json()
        working_set.add(pkg_resources.Distribution(
            metadata=MockProvider({
                'requires.txt': '[dev]\ncalmpy.pip',
            }),
            project_name='site',
            version='0.0.0',
        ))
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
        )
        result = driver.pkg_manager_view('site')
        self.assertEqual(result, {
            "require": {},
            "name": "site",
        })
        result = driver.pkg_manager_view('site[dev]')
        self.assertEqual(result, {
            "require": {"setuptools": "25.1.6"},
            # should be "site[dev]", but npm fails on that.
            "name": "site",
        })

    def test_pkg_manager_view_bad_entry_point(self):
        self.setup_requirements_json()
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',),
        )
        with self.assertRaises(ValueError) as e:
            driver.pkg_manager_view('calmpy.pip [dev]')
        self.assertIn(
            'malformed package name(s) specified: [dev]',
            str(e.exception))

        with self.assertRaises(ValueError) as e:
            driver.pkg_manager_view('{foo} /r')
        self.assertIn(
            'malformed package name(s) specified: {foo}, /r',
            str(e.exception))

    def test_pkg_manager_cmd_prodev_flag_basic(self):
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr')
        with pretty_logging(stream=mocks.StringIO()) as log:
            # production has priority
            self.assertEqual(driver._prodev_flag(True, True, False), [
                '--production=true'])
            self.assertEqual(driver._prodev_flag(False, True, False), [
                '--production=false'])
            self.assertEqual(driver._prodev_flag(True, False, True), [
                '--production=true'])
            self.assertEqual(driver._prodev_flag(False, False, True), [
                '--production=false'])
        self.assertEqual(log.getvalue(), '')

        with pretty_logging(stream=mocks.StringIO()) as log:
            self.assertEqual(driver._prodev_flag(True, None, False), [
                '--production=true'])
            self.assertEqual(driver._prodev_flag(False, None, False), [
                '--production=false'])
            self.assertEqual(driver._prodev_flag(True, None, True), [
                '--production=true'])
            self.assertEqual(driver._prodev_flag(False, None, True), [
                '--production=false'])
        self.assertEqual(log.getvalue(), '')

        with pretty_logging(stream=mocks.StringIO()) as log:
            self.assertEqual(driver._prodev_flag(None, True, False), [
                '--production=false'])
            self.assertEqual(driver._prodev_flag(None, False, False), [
                '--production=true'])
            self.assertEqual(driver._prodev_flag(None, True, True), [
                '--production=false'])
            self.assertEqual(driver._prodev_flag(None, False, True), [
                '--production=true'])
        self.assertEqual(log.getvalue(), '')

        with pretty_logging(stream=mocks.StringIO()) as log:
            self.assertEqual(driver._prodev_flag(None, None, False), [])
        self.assertNotIn('WARNING', log.getvalue())
        self.assertIn('DEBUG', log.getvalue())
        self.assertIn(
            "no packages defined in 'devDependencies' section", log.getvalue())

    def test_pkg_manager_cmd_production_flag_warnings_interactive(self):
        stub_check_interactive(self, True)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr', devkey='blah')
        with pretty_logging(stream=mocks.StringIO()) as log:
            self.assertEqual(driver._prodev_flag(None, None, True), [])
        self.assertIn('WARNING', log.getvalue())
        self.assertIn(
            'undefined production flag may result in unexpected installation '
            'behavior', log.getvalue()
        )
        self.assertNotIn("non-interactive", log.getvalue())
        self.assertIn("'blah' may be installed", log.getvalue())

    def test_pkg_manager_cmd_production_flag_warnings_noninteractive(self):
        stub_check_interactive(self, False)
        driver = cli.PackageManagerDriver(pkg_manager_bin='mgr', devkey='blah')
        with pretty_logging(stream=mocks.StringIO()) as log:
            self.assertEqual(driver._prodev_flag(None, None, True), [])
        self.assertIn('WARNING', log.getvalue())
        self.assertIn(
            'undefined production flag may result in unexpected installation '
            'behavior', log.getvalue()
        )
        self.assertIn("non-interactive", log.getvalue())
        self.assertIn("'blah' may be ignored", log.getvalue())

    def test_pkg_manager_cmd_production_flag_unset(self):
        stub_check_interactive(self, False)
        stub_mod_call(self, cli)
        stub_base_which(self)
        self.setup_requirements_json()
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',), devkey='require',
        )

        with pretty_logging(stream=mocks.StringIO()) as log:
            driver.pkg_manager_install(['calmpy.pip'])

        self.assertIn('WARNING', log.getvalue())
        self.assertIn(
            'undefined production flag may result in unexpected installation '
            'behavior', log.getvalue()
        )
        self.assertIn("non-interactive", log.getvalue())
        self.assertIn("'require' may be ignored", log.getvalue())
        self.assertEqual(self.call_args[0], (['mgr', 'install'],))

    def test_pkg_manager_cmd_production_flag_set(self):
        stub_mod_call(self, cli)
        stub_base_which(self)
        self.setup_requirements_json()
        driver = cli.PackageManagerDriver(
            pkg_manager_bin='mgr', pkgdef_filename='requirements.json',
            dep_keys=('require',), devkey='require',
        )

        with pretty_logging(stream=mocks.StringIO()) as log:
            driver.pkg_manager_install(['calmpy.pip'], production=True)
        self.assertNotIn('WARNING', log.getvalue())
        self.assertEqual(self.call_args[0], (
            ['mgr', 'install', '--production=true'],))

        with pretty_logging(stream=mocks.StringIO()) as log:
            driver.pkg_manager_install(['calmpy.pip'], production=False)
        self.assertNotIn('WARNING', log.getvalue())
        self.assertEqual(self.call_args[0], (
            ['mgr', 'install', '--production=false'],))



# -*- coding: utf-8 -*-
"""
Assortment of utility functions.
"""

from __future__ import absolute_import

import logging
import os
import re
import sys
from contextlib import contextmanager
from functools import partial
from json import dump
from json import dumps
from locale import getpreferredencoding
from os import strerror
from os.path import curdir
from os.path import defpath
from os.path import normcase
from os.path import pathsep
from pdb import post_mortem
from pdb import Pdb
from subprocess import Popen
from subprocess import PIPE

locale = getpreferredencoding()

# sys.platform have required keys for environment variables for Popen
_PLATFORM_ENV_KEYS = {
    # win32 specific keys
    # APPDATA
    #     Some node scripts really need this for some reason?
    # PATH
    #     Well, if it's not available, funnily enough even with a full
    #     path executable things will just also not work.
    # PATHEXT
    #     For cases where supplied argument has no PATHEXT
    # SYSTEMROOT
    #     Without this, on Windows 7 (probably others) will just result
    #     in "socket: (10106)" error.
    'win32': ['APPDATA', 'PATH', 'PATHEXT', 'SYSTEMROOT'],
}

# split comma via negative lookahead - match all things inside [] which
# includes the `,` that may be inside that to avoid splitting.
requirement_comma_list = re.compile(',(?![^\\[\\]]*\\])')

json_dumps = partial(dumps, indent=4, sort_keys=True, separators=(',', ': '))
json_dump = partial(dump, indent=4, sort_keys=True, separators=(',', ': '))


def enable_pretty_logging(logger='calmjs', level=logging.DEBUG, stream=None):
    """
    Shorthand to enable pretty logging
    """

    def cleanup():
        logger.removeHandler(handler)
        logger.level = old_level

    if not isinstance(logger, logging.Logger):
        logger = logging.getLogger(logger)

    old_level = logger.level
    handler = logging.StreamHandler(stream)
    handler.setFormatter(logging.Formatter(
        u'%(asctime)s %(levelname)s %(name)s %(message)s'))
    logger.addHandler(handler)
    logger.setLevel(level)
    return cleanup


@contextmanager
def pretty_logging(logger='calmjs', level=logging.DEBUG, stream=None):
    try:
        cleanup = enable_pretty_logging(logger, level, stream)
        yield stream
    finally:
        cleanup()


def finalize_env(env):
    """
    Produce a platform specific env for passing into subprocess.Popen
    family of external process calling methods, and the supplied env
    will be updated on top of it.  Returns a new env.
    """

    keys = _PLATFORM_ENV_KEYS.get(sys.platform, [])
    if 'PATH' not in keys:
        # this MUST be available due to Node.js (and others really)
        # needing something to look for binary locations when it shells
        # out to other binaries.
        keys.append('PATH')
    results = {
        key: os.environ.get(key, '') for key in keys
    }
    results.update(env)
    return results


def fork_exec(args, stdin='', **kwargs):
    """
    Do a fork-exec through the subprocess.Popen abstraction in a way
    that takes a stdin and return stdout.
    """

    as_bytes = isinstance(stdin, bytes)
    source = stdin if as_bytes else stdin.encode(locale)
    p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, **kwargs)
    stdout, stderr = p.communicate(source)
    if as_bytes:
        return stdout, stderr
    return (stdout.decode(locale), stderr.decode(locale))


def raise_os_error(_errno, path=None):
    """
    Helper for raising the correct exception under Python 3 while still
    being able to raise the same common exception class in Python 2.7.
    """

    msg = "%s: '%s'" % (strerror(_errno), path) if path else strerror(_errno)
    raise OSError(_errno, msg)


def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """
    Given cmd, check where it is on PATH.

    Loosely based on the version in python 3.3.
    """

    if os.path.dirname(cmd):
        if os.path.isfile(cmd) and os.access(cmd, mode):
            return cmd

    if path is None:
        path = os.environ.get('PATH', defpath)
    if not path:
        return None

    paths = path.split(pathsep)

    if sys.platform == 'win32':
        # oh boy
        if curdir not in paths:
            paths = [curdir] + paths

        # also need to check the fileexts...
        pathext = os.environ.get('PATHEXT', '').split(pathsep)

        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # sanity
        files = [cmd]

    seen = set()
    for p in paths:
        normpath = normcase(p)
        if normpath in seen:
            continue
        seen.add(normpath)
        for f in files:
            fn = os.path.join(p, f)
            if os.path.isfile(fn) and os.access(fn, mode):
                return fn

    return None


def pdb_post_mortem(*a, **kw):
    post_mortem(*a, **kw)


def pdb_set_trace(*a, **kw):
    Pdb(skip=[__name__], *a, **kw).set_trace()



# -*- coding: utf-8 -*-
"""
Registry of all registries for calmjs.

This module provides the implementation for working with the entry point
group ``calmjs.registry``, where it will be able to instantiate and
cache registries for the rest of the calmjs framework.

Typically, usage is done through this::

    >>> from calmjs.registry import get
    >>> calmjs_module = get('calmjs.module')
    >>> calmjs_module  # doctest: +ELLIPSIS
    <calmjs.module.ModuleRegistry object ...>
    >>> calmjs_module.iter_records()  # doctest: +ELLIPSIS
    <generator object ...>

The iter_records is typically used by the other parts of the calmjs
framework to produce configuration files and/or transpile the source
into the usable final form.
"""

from __future__ import absolute_import

from pkg_resources import working_set
from pkg_resources import Requirement

from logging import getLogger
from calmjs.base import BaseRegistry

logger = getLogger(__name__)

PACKAGE_NAME = 'calmjs'
CALMJS_RESERVED = 'calmjs.reserved'


class Registry(BaseRegistry):

    def __init__(
            self, registry_name,
            package_name=PACKAGE_NAME, reserved=CALMJS_RESERVED, *a, **kw):
        """
        The root registry will also track the undocumented working set
        specification
        """

        working_set_ = kw.get('_working_set') or working_set
        self.reserved = {}
        if reserved:
            dist = working_set_.find(Requirement.parse(package_name))
            if dist is None:
                logger.error(
                    "failed to set up registry_name reservations for "
                    "registry '%s', as the specified package '%s' could "
                    "not found in the current working_set; maybe it is not "
                    "correctly installed?", registry_name, package_name,
                )
            else:
                # module_name is the package_name in our context.
                self.reserved = {
                    k: v.module_name for k, v in dist.get_entry_map(
                        reserved).items()
                }
        super(Registry, self).__init__(registry_name, *a, **kw)

    def _init(self):
        """
        Turn the records into actual usable keys.
        """

        self._entry_points = {}
        for entry_point in self.raw_entry_points:
            if entry_point.dist.project_name != self.reserved.get(
                    entry_point.name, entry_point.dist.project_name):
                logger.error(
                    "registry '%s' for '%s' is reserved for package '%s'",
                    entry_point.name, self.registry_name,
                    self.reserved[entry_point.name],
                )
                continue

            if self.get_record(entry_point.name):
                logger.warning(
                    "registry '%s' for '%s' is already registered.",
                    entry_point.name, self.registry_name,
                )
                existing = self._entry_points[entry_point.name]
                logger.debug(
                    "registered '%s' from '%s'", existing, existing.dist)
                logger.debug(
                    "discarded '%s' from '%s'", entry_point, entry_point.dist)
                continue

            logger.debug(
                "recording '%s' from '%s'", entry_point, entry_point.dist)
            self._entry_points[entry_point.name] = entry_point

        # No pre-caching for below, let the get_record load things into
        # records on-demand.

    def get_record(self, name):
        if name in self.records:
            # maybe do some other sanity check if pedantic.
            logger.debug('found existing registry %s', name)
            return self.records[name]

        entry_point = self._entry_points.get(name)
        if not entry_point:
            logger.debug("'%s' does not resolve to a registry", name)
            return

        try:
            cls = entry_point.load()
        except ImportError:
            logger.debug(
                "ImportError '%s' from '%s'",
                entry_point, entry_point.dist)
            return

        if cls is type(self) and entry_point.name == self.registry_name:
            logger.debug(
                "registry '%s' has entry point '%s' which is the identity "
                "registration", name, entry_point,
            )
            self.records[name] = self
            return self

        logger.debug(
            "registering '%s' from '%s'", entry_point, entry_point.dist)
        try:
            self.records[name] = cls(name)
        except Exception:
            logger.exception(
                "'%s' from '%s' does not lead to a valid registry constructor",
                entry_point, entry_point.dist,
            )
            return
        return self.records[name]


# Initialize the root registry instance
_inst = Registry(__name__)  # __name__ == calmjs.registry


def get(registry_name):
    return _inst.get(registry_name)



from inspect import currentframe
from inspect import getouterframes

from calmjs import runtime


class BadSimpleRuntime(runtime.DriverRuntime, runtime.Runtime):

    def entry_point_load_validated(self, entry_point):
        # skip the rest of the checks.
        return entry_point.load()

    def init_argparser(self, argparser):
        level = len(getouterframes(currentframe()))
        if level > self.recursionlimit:
            # turns out we need to emulate this to make pypy not
            # blow up coverage reporting; also make it die
            # quicker, and this emulation works good enough as
            # it turns out.
            raise RuntimeError('maximum recursion depth exceeded')
        super(BadSimpleRuntime, self).init_argparser(argparser)


class FakeBootstrapRuntime(runtime.BootstrapRuntime):
    pass


fake_bootstrap = FakeBootstrapRuntime()



# -*- coding: utf-8 -*-
import unittest
from pkg_resources import Distribution
from pkg_resources import EntryPoint

import calmjs.base
from calmjs.base import BaseModuleRegistry
from calmjs.registry import Registry
from calmjs.registry import get
from calmjs.module import ModuleRegistry
from calmjs.module import PythonicModuleRegistry
from calmjs.module import resolve_child_module_registries_lineage
from calmjs.utils import pretty_logging

from calmjs.testing import mocks
from calmjs.testing import utils
from calmjs.testing.module import ChildModuleRegistry


class ModuleRegistryTestCase(unittest.TestCase):
    """
    Test the JavaScript module registry.
    """

    def setUp(self):
        self.registry = ModuleRegistry(__name__)

    def test_module_registry_empty(self):
        with self.assertRaises(StopIteration):
            next(self.registry.iter_records())

    def test_module_registry_standard(self):
        with pretty_logging(stream=mocks.StringIO()):
            self.registry.register_entry_points([EntryPoint.parse(
                'calmjs.testing.module1 = calmjs.testing.module1')])
        self.assertEqual(sorted(
            key for key, value in self.registry.iter_records()
        ), [
            'calmjs.testing.module1',
        ])

        module1 = self.registry.get_record('calmjs.testing.module1')
        key = 'calmjs/testing/module1/hello'
        self.assertEqual(sorted(module1.keys()), [key])

    def test_module_registry_pythonic(self):
        registry = PythonicModuleRegistry(__name__)
        with pretty_logging(stream=mocks.StringIO()):
            registry.register_entry_points([EntryPoint.parse(
                'calmjs.testing.module1 = calmjs.testing.module1')])
        self.assertEqual(sorted(
            key for key, value in registry.iter_records()
        ), [
            'calmjs.testing.module1',
        ])

        module1 = registry.get_record('calmjs.testing.module1')
        key = 'calmjs.testing.module1.hello'
        self.assertEqual(sorted(module1.keys()), [key])


class IntegratedModuleRegistryTestCase(unittest.TestCase):
    """
    Test the JavaScript module registry, with a mocked working set and
    going through the root registry.
    """

    def test_module_registry_through_registry(self):
        """
        Show that the module registry instantiated through the global
        registry's get function will result in the resulting module
        registry being populated properly with the module entries when
        using a constrained set of entry points.
        """

        working_set = mocks.WorkingSet({
            'calmjs.module': [
                'module1 = calmjs.testing.module1',
                'module2 = calmjs.testing.module2',
                'module3 = calmjs.testing.module3',
            ],
            'calmjs.reserved': [
                'calmjs.module = calmjs.testing',
            ],
            __name__: [
                'calmjs.module = calmjs.module:ModuleRegistry',
            ]},
            dist=Distribution(project_name='calmjs.testing', version='0.0')
        )
        utils.stub_mod_working_set(self, [calmjs.base], working_set)

        # Not going to use the global registry, and using our custom
        # reservation entry
        local_root_registry = Registry(
            __name__, 'calmjs.testing', _working_set=working_set)
        with pretty_logging(stream=mocks.StringIO()):
            # silences "distribution 'calmjs.testing 0.0' not found"
            # warnings from stdout produced by the indexer, as the
            # provided working_set is invalid with entry points that do
            # not have a valid distribution.
            global_registry = get('calmjs.module')
            registry = local_root_registry.get_record('calmjs.module')
        self.assertIsNot(registry, global_registry)
        self.assertEqual(
            sorted(k for k, v in registry.iter_records()), [
                'calmjs.testing.module1',
                'calmjs.testing.module2',
                'calmjs.testing.module3',
            ]
        )

        results = registry.get_records_for_package('calmjs.testing')
        self.assertEqual(sorted(results.keys()), [
           'calmjs/testing/module1/hello',
           'calmjs/testing/module2/helper',
           'calmjs/testing/module2/index',
           'calmjs/testing/module3/math',
        ])

        module1 = registry.get_record('calmjs.testing.module1')
        self.assertIn('calmjs/testing/module1/hello', module1)


class ExtraJsonKeysRegistryTestCase(unittest.TestCase):

    def test_integrated(self):
        """
        Since this module already immediately declares that, this can
        simply be a quick integration test.
        """

        reg = get('calmjs.extras_keys')
        results = set(reg.iter_records())
        self.assertIn('node_modules', results)


class LineageResolutionTestCase(unittest.TestCase):

    def test_resolve_child_module_registries_lineage_solo(self):
        working_set = mocks.WorkingSet({})
        root = BaseModuleRegistry('root.module', _working_set=working_set)
        it = resolve_child_module_registries_lineage(root)
        self.assertEqual(root, next(it))
        self.assertEqual([], list(it))

    def test_resolve_child_module_registries_lineage_base(self):
        working_set = mocks.WorkingSet({})
        root = BaseModuleRegistry(
            'root.module', _working_set=working_set)
        child = ChildModuleRegistry(
            'root.module.child', _parent=root, _working_set=working_set)
        self.assertEqual([
            root,
            child,
        ], list(resolve_child_module_registries_lineage(child)))

    def test_resolve_child_module_registries_lineage_two_layers(self):
        working_set = mocks.WorkingSet({})
        root = BaseModuleRegistry(
            'root.module', _working_set=working_set)
        child = ChildModuleRegistry(
            'root.module.child', _parent=root, _working_set=working_set)
        grandchild = ChildModuleRegistry(
            'root.module.child.child', _parent=child,
            _working_set=working_set)
        greatgrandchild = ChildModuleRegistry(
            'root.module.child.child.child', _parent=grandchild,
            _working_set=working_set)
        self.assertEqual([
            root,
            child,
            grandchild,
        ], list(resolve_child_module_registries_lineage(grandchild)))
        self.assertEqual([
            root,
            child,
            grandchild,
            greatgrandchild,
        ], list(resolve_child_module_registries_lineage(
            greatgrandchild)))

    def test_resolve_child_module_registries_lineage_malformed_loop(self):
        working_set = mocks.WorkingSet({})
        root = BaseModuleRegistry(
            'root', _working_set=working_set)
        parent = ChildModuleRegistry(
            'root.child', _parent=root, _working_set=working_set)
        child = ChildModuleRegistry(
            'root.child.child', _parent=parent, _working_set=working_set)
        grandchild = ChildModuleRegistry(
            'root.child.child.child', _parent=child, _working_set=working_set)
        # force a bad loop
        parent.parent = grandchild

        with pretty_logging(stream=mocks.StringIO()) as log:
            with self.assertRaises(TypeError) as e:
                resolve_child_module_registries_lineage(parent)

        self.assertEqual(
            "registry 'root.child' was already recorded in the lineage, "
            "indicating that it may be some (grand)child of itself, which is "
            "an illegal reference in the registry system; previously resolved "
            "lineage is: ['root.child.child', 'root.child.child.child', "
            "'root.child']", str(e.exception))
        self.assertIn(
            "the parent registry 'root.child.child.child' somehow has a "
            "longer name than its child registry 'root.child'; the underlying "
            "registry class may be constructed in an invalid manner",
            log.getvalue()
        )

    def test_resolve_child_module_registries_lineage_mismatched_prefixes(self):
        working_set = mocks.WorkingSet({})
        root = BaseModuleRegistry(
            'root', _working_set=working_set)
        parent = ChildModuleRegistry(
            'root.child', _parent=root, _working_set=working_set)
        child = ChildModuleRegistry(
            'root_child.child', _parent=parent, _working_set=working_set)

        with pretty_logging(stream=mocks.StringIO()) as log:
            it = resolve_child_module_registries_lineage(child)

        self.assertIn(
            "child registry 'root_child.child' does not share the same common "
            "prefix as its parent registry 'root.child'; there may be errors "
            "with how the related registries are set up or constructed",
            log.getvalue()
        )
        self.assertEqual([
            'root', 'root.child', 'root_child.child'
        ], [r.registry_name for r in it])



# -*- coding: utf-8 -*-
import io

from pkg_resources import Distribution
from pkg_resources import EntryPoint
from pkg_resources import EmptyProvider

from setuptools.command.egg_info import egg_info

_dist = Distribution()
dummy = object()


class WorkingSet(object):

    def __init__(self, items, dist=_dist):
        self.items = items
        self.dist = dist

    def iter_entry_points(self, name):
        items = self.items.get(name, [])
        for item in items:
            entry_point = item if isinstance(
                item, EntryPoint) else EntryPoint.parse(item)
            entry_point.dist = self.dist
            yield entry_point

    def find(self, name):
        # just always return the dist.
        return self.dist


class Mock_egg_info(egg_info):

    def initialize_options(self):
        egg_info.initialize_options(self)
        self.called = {}

    def write_or_delete_file(self, what, filename, data, force=True):
        """
        Stub out the actual called function
        """

        self.called[filename] = data


class MockProvider(EmptyProvider):
    """
    Extends upon the emptiness of that.
    """

    def __init__(self, metadata):
        self._metadata = {}
        self._metadata.update(metadata)

    def has_metadata(self, name):
        return name in self._metadata

    def get_metadata(self, name):
        results = self._metadata.get(name)
        if results is None:
            raise IOError('emulating an IOError')
        return results


class StringIO(io.StringIO):
    """
    A "safely" wrapped version
    """

    def write(self, msg):
        io.StringIO.write(self, msg.encode(
            'utf8', 'backslashreplace').decode('utf8'))



# -*- coding: utf-8 -*-
"""
Various loaders.
"""

from __future__ import absolute_import

import fnmatch
import pkg_resources

from logging import getLogger
from glob import iglob
from os.path import exists
from os.path import join
from os.path import relpath
from os.path import sep
from os import walk

logger = getLogger(__name__)

JS_EXT = '.js'

_utils = {
    'modpath': {},
    'globber': {},
    'modname': {},
    'mapper': {},
}


def resource_filename_mod_dist(module_name, dist):
    """
    Given a module name and a distribution, attempt to resolve the
    actual path to the module.
    """

    try:
        return pkg_resources.resource_filename(
            dist.as_requirement(), join(*module_name.split('.')))
    except pkg_resources.DistributionNotFound:
        logger.warning(
            "distribution '%s' not found, falling back to resolution using "
            "module_name '%s'", dist, module_name,
        )
        return pkg_resources.resource_filename(module_name, '')


# An attempt was made to use the provided distribution argument directly
# implemented as following:
#
# def resource_filename_mod_dist(module_name, dist):
#     """
#     Given a module name and a distribution, with the assumption that the
#     distribution is part of the default working set, resolve the actual
#     path to the module.
#     """
#
#     return dist.get_resource_filename(
#         pkg_resources.working_set, join(*module_name.split('.')))
#

# However, we cannot necessary access the underlying resource manager as
# that is "hidden" as an implementation detail, and that this duplicates
# code provided by that class.  Also due to the inconsistency with how
# values are handled, namely that the provided manager is generally
# irrelvant unless the provided distribution belongs to/references a
# zipped-egg, which IProvider.get_resource_filename implementation will
# actually make use of the manager to acquire relevant information.
#
# That said, just for clarity (and not waste some hours that was spent
# investigating whether we can bypass certain things to save some time
# on other test setup), it was discovered that the resolution of the
# path is done during the construction of the working_set, such that it
# ultimately will result in a path if the dist.as_requirement is not
# called and that the get_resource_filename is invoked directly.  The
# points of interests in the pkg_resources module:
# - find_on_path (resolves below via dist_factory)
# - distributions_from_metadata (which uses PathMetadata)
# The relevant path for the Distribution is ultimately found in
# `Distribution._provider.module_path`.


def resource_filename_mod_entry_point(module_name, entry_point):
    """
    If a given package declares a namespace and also provide submodules
    nested at that namespace level, and for whatever reason that module
    is needed, Python's import mechanism will not have a path associated
    with that module.  However, if given an entry_point, this path can
    be resolved through its distribution.  That said, the default
    resource_filename function does not accept an entry_point, and so we
    have to chain that back together manually.
    """

    if entry_point.dist is None:
        # distribution missing is typically caused by mocked entry
        # points from tests; silently falling back to basic lookup
        result = pkg_resources.resource_filename(module_name, '')
    else:
        result = resource_filename_mod_dist(module_name, entry_point.dist)

    if not result:
        logger.warning(
            "fail to resolve the resource path for module '%s' and "
            "entry_point '%s'", module_name, entry_point
        )
        return None
    if not exists(result):
        logger.warning(
            "resource path resolved to be '%s' for module '%s' and "
            "entry_point '%s', but it does not exist",
            result, module_name, entry_point,
        )
        return None
    return result


def modgen(
        module, entry_point,
        modpath='pkg_resources', globber='root', fext=JS_EXT,
        registry=_utils):
    """
    JavaScript styled module location listing generator.

    Arguments:

    module
        The Python module to start fetching from.

    entry_point
        This is the original entry point that has a distribution
        reference such that the resource_filename API call may be used
        to locate the actual resources.

    Optional Arguments:

    modpath
        The name to the registered modpath function that will fetch the
        paths belonging to the module.  Defaults to 'pkg_resources'.

    globber
        The name to the registered file globbing function.  Defaults to
        one that will only glob the local path.

    fext
        The filename extension to match.  Defaults to `.js`.

    registry
        The "registry" to extract the functions from

    Yields 3-tuples of

    - raw list of module name fragments
    - the source base path to the python module (equivalent to module)
    - the relative path to the actual module

    For each of the module basepath and source files the globber finds.
    """

    globber_f = globber if callable(globber) else registry['globber'][globber]
    modpath_f = modpath if callable(modpath) else registry['modpath'][modpath]

    logger.debug(
        'modgen generating file listing for module %s',
        module.__name__,
    )

    module_frags = module.__name__.split('.')
    module_base_paths = modpath_f(module, entry_point)

    for module_base_path in module_base_paths:
        logger.debug('searching for *%s files in %s', fext, module_base_path)
        for path in globber_f(module_base_path, '*' + fext):
            mod_path = (relpath(path, module_base_path))
            yield (
                module_frags + mod_path[:-len(fext)].split(sep),
                module_base_path,
                mod_path,
            )


def register(util_type, registry=_utils):
    """
    Crude, local registration decorator for a crude local registry of
    all utilities local to this module.
    """

    def marker(f):
        mark = util_type + '_'
        if not f.__name__.startswith(mark):
            raise TypeError(
                'not registering %s to %s' % (f.__name__, util_type))
        registry[util_type][f.__name__[len(mark):]] = f
        return f
    return marker


@register('modpath')
def modpath_all(module, entry_point):
    """
    Provides the raw __path__.  Incompatible with PEP 302-based import
    hooks and incompatible with zip_safe packages.

    Deprecated.  Will be removed by calmjs-4.0.
    """

    module_paths = getattr(module, '__path__', [])
    if not module_paths:
        logger.warning(
            "module '%s' does not appear to be a namespace module or does not "
            "export available paths onto the filesystem; JavaScript source "
            "files cannot be extracted from this module.", module.__name__
        )
    return module_paths


@register('modpath')
def modpath_last(module, entry_point):
    """
    Provides the raw __path__.  Incompatible with PEP 302-based import
    hooks and incompatible with zip_safe packages.

    Deprecated.  Will be removed by calmjs-4.0.
    """

    module_paths = modpath_all(module, entry_point)
    if len(module_paths) > 1:
        logger.info(
            "module '%s' has multiple paths, default selecting '%s' as base.",
            module.__name__, module_paths[-1],
        )
    return module_paths[-1:]


@register('modpath')
def modpath_pkg_resources(module, entry_point):
    """
    Goes through pkg_resources for compliance with various PEPs.

    This one accepts a module as argument.
    """

    result = []
    try:
        path = resource_filename_mod_entry_point(module.__name__, entry_point)
    except ImportError:
        logger.warning("module '%s' could not be imported", module.__name__)
    except Exception:
        logger.warning("%r does not appear to be a valid module", module)
    else:
        if path:
            result.append(path)
    return result


@register('globber')
def globber_root(root, patt):
    return iglob(join(root, patt))


@register('globber')
def globber_recursive(root, patt):
    for root, dirnames, filenames in walk(root):
        for filename in fnmatch.filter(filenames, patt):
            yield join(root, filename)


@register('modname')
def modname_es6(fragments):
    """
    Generates ES6 styled module names from fragments.
    """

    return '/'.join(fragments)


@register('modname')
def modname_python(fragments):
    """
    Generates Python styled module names from fragments.
    """

    return '.'.join(fragments)


def mapper(module, entry_point,
           modpath='pkg_resources', globber='root', modname='es6',
           fext=JS_EXT, registry=_utils):
    """
    General mapper

    Loads components from the micro registry.
    """

    modname_f = modname if callable(modname) else _utils['modname'][modname]

    return {
        modname_f(modname_fragments): join(base, subpath)
        for modname_fragments, base, subpath in modgen(
            module, entry_point=entry_point,
            modpath=modpath, globber=globber,
            fext=fext, registry=_utils)
    }


@register('mapper')
def mapper_es6(module, entry_point, globber='root', fext=JS_EXT):
    """
    Default mapper

    Finds the latest path declared for the module at hand and extract
    a list of importable JS modules using the es6 module import format.
    """

    return mapper(
        module, entry_point=entry_point, modpath='pkg_resources',
        globber=globber, modname='es6', fext=fext)


@register('mapper')
def mapper_python(module, entry_point, globber='root', fext=JS_EXT):
    """
    Default mapper using python style globber

    Finds the latest path declared for the module at hand and extract
    a list of importable JS modules using the es6 module import format.
    """

    return mapper(
        module, entry_point=entry_point, modpath='pkg_resources',
        globber=globber, modname='python', fext=fext)



# -*- coding: utf-8 -*-
"""
Management of the production and querying of prebuilt artifacts.

The design for the artifact registry system makes a number of tradeoffs
to increase end-user usability and simplicity, by not using another
registry to manage the sets of common functions for artifact production
that are designated to be compatible.  The usage of the
``EntryPoint.attrs`` as the authoritative identifier for the group of
common functions without delegating this role to another registry has
the benefit of simplifying lookup, but this means that any future
changes to the meaning of the name of that group should not be done as
the versioning of this can become quite complicated.

This also mean that the usage of extra metadata is required for
recording which version of what package actually resulted in the
construction of that package.  Also, this means the various toolchain
packages must cooperate on what the meaning of the words are, in order
for a single artifact registry instance to function in an unambiguous
manner.

To achieve this, the functions registered to the registry MUST accept
the two required arguments, ``package_names`` and ``export_target`` (for
the production of the ``Spec`` instance), and MUST NOT produce the
artifacts directly, but instead return a 2-tuple containing a
``Toolchain`` and ``Spec`` instance.  This return value is then invoked
by the build method within the ``ArtifactRegistry`` instance which will
complete the build task.  The results can be queried through the methods
offered by the registry.

Correct functionality, of course, depends fully on the package and the
developer who built the artifact, and requires that the process as
outlined in the API is fully followed without artificial out-of-band
manipulation, as there are zero enforcement options for guarding against
what actually goes into a wheel.

For example, given the two following packages that provide a calmjs
toolchain implementation:

- default.gloop
- default.glump

They could provide a module that contain functions such as:

- default.gloop.build:gloop_artifact
- default.glump.build:glump_artifact

A package that require artifacts built, ``example.package``, may have
this declaration in its entry points:

    [calmjs.artifacts]
    deploy.gloop.js = default.gloop.build:gloop_artifact
    deploy.glump.js = default.glump.build:glump_artifact

The build process for both artifacts can be manually trigger by the
following code:

    >>> from calmjs.registry import get
    >>> artifacts = get('calmjs.artifacts')
    >>> artifacts.process_package('example.package')
    Building artifact for deploy.gloop.js
    Building artifact for deploy.glump.js

When the build process is done, inside the egg-info or dist-info
directory of the ``example.package`` package will have a new
``calmjs_artifacts`` directory that contain both those new artifact
files.

Resolution for the location of the artifact can be done using the
``resolve_artifacts_by_builder_compat`` method on the registry instance,
for example:

    >>> list(artifacts.resolve_artifacts_by_builder_compat(
    ...     'example.package', 'gloop_artifact'))
    ['/.../src/example.package.egg-info/calmjs_artifacts/deploy.gloop.js']

However, if the ``example.package`` wish to have its own build process
for either of those artifacts, it can declare a new builder with the
same name in itself like so:

- example.package.builder:gloop_artifact

Thus changing the entry point to this:

    [calmjs.artifacts]
    deploy.gloop.js = example.package.builder:gloop_artifact
    deploy.glump.js = default.glump.builder:glump_artifact

The resolve method should bring up the information.

To enable the building of artifacts while building the package using
setuptools, in the ``setup`` function (typically of ``setup.py``) the
``build_calmjs_artifacts` should be set to True.  Combining together
with the entry point setup, the setup call may look something like this:

    setup(
        name='example.package',
        # ...
        build_calmjs_artifacts=True,
        entry_points={
            # module declaration
            'calmjs.module': [
                'example.package = example.package',
            ],
            'calmjs.artifacts': [
                'deploy.gloop.js = example.package.builder:gloop_artifact',
                'deploy.glump.js = default.glump.builder:glump_artifact',
            ],
        },
        build_calmjs_artifacts=True,
        # ...
    )
"""

from __future__ import absolute_import

import json
import warnings
from codecs import open
from inspect import getcallargs
from inspect import getmro
from logging import getLogger
from os.path import basename
from os.path import dirname
from os.path import exists
from os.path import isdir
from os.path import join
from os.path import normcase
from os import makedirs
from os import unlink
from shutil import rmtree

from calmjs.base import BaseRegistry
from calmjs.base import PackageKeyMapping
from calmjs.dist import find_packages_requirements_dists
from calmjs.dist import find_pkg_dist
from calmjs.dist import is_json_compat
from calmjs.dist import pkg_names_to_dists
from calmjs.cli import get_bin_version_str
from calmjs.command import BuildArtifactCommand
from calmjs.registry import get
from calmjs.types.exceptions import ToolchainAbort
from calmjs.toolchain import Toolchain
from calmjs.toolchain import Spec
from calmjs.toolchain import TOOLCHAIN_BIN_PATH
from calmjs.toolchain import BEFORE_PREPARE
from calmjs.toolchain import EXPORT_TARGET

ARTIFACT_BASENAME = 'calmjs_artifacts'
ARTIFACT_REGISTRY_NAME = 'calmjs.artifacts'

logger = getLogger(__name__)


def _cls_lookup_dist(cls):
    """
    Attempt to resolve the distribution from the provided class in the
    most naive way - this assumes the Python module path to the class
    contains the name of the package that provided the module and class.
    """

    frags = cls.__module__.split('.')
    for name in ('.'.join(frags[:x]) for x in range(len(frags), 0, -1)):
        dist = find_pkg_dist(name)
        if dist:
            return dist


def verify_builder(builder):
    """
    To ensure that the provided builder has a signature that is at least
    compatible.
    """

    try:
        d = getcallargs(builder, package_names=[], export_target='some_path')
    except TypeError:
        return False
    return d == {'package_names': [], 'export_target': 'some_path'}


def extract_builder_result(builder_result, toolchain_cls=Toolchain):
    """
    Extract the builder result to produce a ``Toolchain`` and ``Spec``
    instance.
    """

    try:
        toolchain, spec = builder_result
    except Exception:
        return None, None
    if not isinstance(toolchain, toolchain_cls) or not isinstance(spec, Spec):
        return None, None
    return toolchain, spec


def trace_toolchain(toolchain):
    """
    Trace the versions of the involved packages for the provided
    toolchain instance.
    """

    pkgs = []
    for cls in getmro(type(toolchain)):
        if not issubclass(cls, Toolchain):
            continue
        dist = _cls_lookup_dist(cls)
        value = {
            'project_name': dist.project_name,
            'version': dist.version,
        } if dist else {}
        key = '%s:%s' % (cls.__module__, cls.__name__)
        pkgs.append({key: value})
    return pkgs


def setup_export_location(export_target):
    target_dir = dirname(export_target)
    try:
        if not exists(target_dir):
            makedirs(target_dir)
            logger.debug("created '%s' for '%s'", target_dir, export_target)
        elif not isdir(target_dir):
            logger.error(
                "cannot export to '%s' as its dirname does not lead to a "
                "directory", export_target
            )
            return False
        elif isdir(export_target):
            logger.debug(
                "removing existing export target directory at '%s'",
                export_target
            )
            rmtree(export_target)
        elif exists(export_target):
            logger.debug(
                "removing existing export target at '%s'", export_target)
            unlink(export_target)
    except (IOError, OSError) as e:
        logger.error(
            "failed to prepare export location '%s': %s; ensure that any file "
            "permission issues are corrected and/or remove the egg-info "
            "directory for this package before trying again",
            target_dir, e)
        return False

    return True


def prepare_export_location(export_target):
    """
    The version of setup_export_location for use with spec advice
    before_prepare step.
    """

    if not setup_export_location(export_target):
        raise ToolchainAbort()
    return True


class BaseArtifactRegistry(BaseRegistry):
    """
    The base artifact registry implementation.
    A registry to allow a central place for Python packages to declare
    which method to generate the artifact and what name to use.
    """

    def _init(self):
        # default (self.records) is a map of package + name to path
        # this is the reverse lookup for that
        self.reverse = {}
        self.packages = PackageKeyMapping()
        # metadata file about the artifacts for the package
        self.metadata = PackageKeyMapping()
        # for storing builders that are assumed to be compatible due to
        # their identical attribute names.
        self.compat_builders = PackageKeyMapping()
        # TODO determine if the full import name lookup table is
        # required.
        # self.builders = {}

        for ep in self.raw_entry_points:
            # the expected path to the artifact.
            self.register_entry_point(ep)

    def register_entry_point(self, ep):
        path = join(ep.dist.egg_info, ARTIFACT_BASENAME, ep.name)
        nc_path = normcase(path)

        if nc_path in self.reverse:
            logger.error(
                "entry point '%s' from package '%s' resolves to the path '%s' "
                "which was already registered to entry point '%s'; "
                "conflicting entry point registration will be ignored.",
                ep, ep.dist, path, self.reverse[nc_path]
            )
            if normcase(ep.name) != ep.name:
                logger.error(
                    "the file mapping error is caused by this platform's case-"
                    "insensitive filename and the package '%s' defined case-"
                    "sensitive names for the affected artifact"
                )
            return

        # for lookup of the generator(s) for the given package
        p = self.packages[ep.dist.project_name] = self.packages.get(
            ep.dist.project_name, {})
        p[ep.name] = ep

        # for lookup of the path by the builder identified by the
        # compat name (the attrs of the provided entry points) for
        # Python packages.
        cb_key = '.'.join(ep.attrs)
        cb = self.compat_builders[cb_key] = self.compat_builders.get(
            cb_key, PackageKeyMapping())
        cb[ep.dist.project_name] = path

        # for looking up/storage of metadata about the built artifacts.
        self.metadata[ep.dist.project_name] = join(
            ep.dist.egg_info, ARTIFACT_BASENAME + '.json')

        # standard get_artifact_filename lookup for standalone,
        # complete artifacts at some path.
        # TODO consider changing the storage implementation to not use
        # tuple but rather use the PackageKeyMapping for the package
        # name, then a secondary inner lookup for the artifact_name.
        # This is to maintain consistency across implementation
        self.records[(ep.dist.project_name, ep.name)] = path
        # only the reverse lookup must be normalized
        self.reverse[nc_path] = ep

    def iter_records(self):
        # not especially useful, but implementing for completeness.
        for k in self.records.keys():
            yield k

    def belongs_to(self, path):
        """
        Lookup which entry point generated this path.
        """

        return self.reverse.get(normcase(path))

    def get_artifact_filename(self, package_name, artifact_name):
        """
        Similar to pkg_resources.resource_filename, however this works
        with the information cached in this registry instance, and
        arguments are not quite the same.

        Arguments:

        package_name
            The name of the package to get the artifact from
        artifact_name
            The exact name of the artifact.

        Returns the path of where the artifact should be if it has been
        declared, otherwise None.
        """

        project_name = self.packages.normalize(package_name)
        return self.records.get((project_name, artifact_name))

    def resolve_artifacts_by_builder_compat(
            self, package_names, builder_name, dependencies=False):
        """
        Yield the list of paths to the artifacts in the order of the
        dependency resolution

        Arguments:

        package_names
            The names of the packages to probe the dependency graph, to
            be provided as a list of strings.
        artifact_name
            The exact name of the artifact.
        dependencies
            Trace dependencies.  Default is off.

        Returns the path of where the artifact should be if it has been
        declared, otherwise None.
        """

        paths = self.compat_builders.get(builder_name)
        if not paths:
            # perhaps warn, but just return
            return

        resolver = (
            # traces dependencies for distribution.
            find_packages_requirements_dists
            if dependencies else
            # just get grabs the distribution.
            pkg_names_to_dists
        )
        for distribution in resolver(package_names):
            path = paths.get(distribution.project_name)
            if path:
                yield path

    def get_artifact_metadata(self, package_name):
        """
        Return metadata of the artifacts built through this registry.
        """

        filename = self.metadata.get(package_name)
        if not filename or not exists(filename):
            return {}
        with open(filename, encoding='utf8') as fd:
            contents = fd.read()

        try:
            is_json_compat(contents)
        except ValueError:
            logger.info("artifact metadata file '%s' is invalid", filename)
            return {}

        return json.loads(contents)

    def generate_metadata_entry(self, entry_point, toolchain, spec):
        """
        After the toolchain and spec have been executed, this may be
        called to generate the artifact export entry for persistence
        into the metadata file.
        """

        export_target = spec['export_target']
        toolchain_bases = trace_toolchain(toolchain)
        toolchain_bin_path = spec.get(TOOLCHAIN_BIN_PATH)
        toolchain_bin = ([
            basename(toolchain_bin_path),  # bin_name
            get_bin_version_str(toolchain_bin_path),  # bin_version
        ] if toolchain_bin_path else [])

        return {basename(export_target): {
            'toolchain_bases': toolchain_bases,
            'toolchain_bin': toolchain_bin,
            'builder': '%s:%s' % (
                entry_point.module_name, '.'.join(entry_point.attrs)),
        }}

    def update_artifact_metadata(self, package_name, new_metadata):
        metadata = self.get_artifact_metadata(package_name)
        metadata_filename = self.metadata.get(package_name, NotImplemented)
        if metadata_filename is NotImplemented:
            logger.info(
                "package '%s' has not declare any artifacts; not applying "
                "updates to metadata information", package_name)
            return

        artifacts = metadata[ARTIFACT_BASENAME] = metadata.get(
            ARTIFACT_BASENAME, {})
        artifacts.update(new_metadata)
        metadata['versions'] = sorted(set(
            '%s' % i for i in find_packages_requirements_dists(
                [package_name])))

        with open(metadata_filename, 'w', encoding='utf8') as fd:
            json.dump(metadata, fd)

    def iter_records_for(self, package_name):
        """
        Iterate records for a specific package.
        """

        entry_points = self.packages.get(package_name, NotImplemented)
        if entry_points is NotImplemented:
            logger.debug(
                "package '%s' has not declared any entry points for the '%s' "
                "registry for artifact construction",
                package_name, self.registry_name,
            )
            return iter([])

        logger.debug(
            "package '%s' has declared %d entry points for the '%s' "
            "registry for artifact construction",
            package_name, len(entry_points), self.registry_name,
        )
        return iter(entry_points.values())

    def verify_builder(self, builder):
        return verify_builder(builder)

    def verify_export_target(self, export_target):
        # this is the pre-flight verification, it will prevent the
        # entry point loaded/associated with this export_target from
        # creating a builder entry in the iter_builders_for method.
        return True

    def setup_export_location(self, export_target):
        """
        This will be used by prepare_export_location; it must return True
        for the final export to proceed.
        """

        return setup_export_location(export_target)

    def prepare_export_location(self, export_target):
        """
        The wrapped version for this class for use directly by the spec
        in the before_prepare advice.  Must raise some kind of Toolchain
        specific exception to halt Toolchain execution for the given
        export target.
        """

        if not self.setup_export_location(export_target):
            raise ToolchainAbort()

    def extract_builder_result(self, builder_result):
        return extract_builder_result(builder_result)

    def iter_export_targets_for(self, package_name):
        for entry_point in self.iter_records_for(package_name):
            export_target = self.records[
                (entry_point.dist.project_name, entry_point.name)]

            yield entry_point, export_target

    def generate_builder(self, entry_point, export_target):
        """
        Yields exactly one builder if both the provided entry point and
        export target satisfies the checks required.
        """

        try:
            builder = entry_point.resolve()
        except ImportError:
            logger.error(
                "unable to import the target builder for the entry point "
                "'%s' from package '%s' to generate artifact '%s'",
                entry_point, entry_point.dist, export_target,
            )
            return

        if not self.verify_builder(builder):
            logger.error(
                "the builder referenced by the entry point '%s' "
                "from package '%s' has an incompatible signature",
                entry_point, entry_point.dist,
            )
            return

        # CLEANUP see deprecation notice below
        verifier = self.verify_export_target(export_target)
        if not verifier:
            logger.error(
                "the export target '%s' has been rejected", export_target)
            return

        toolchain, spec = self.extract_builder_result(builder(
            [entry_point.dist.project_name], export_target=export_target))
        if not toolchain:
            logger.error(
                "the builder referenced by the entry point '%s' "
                "from package '%s' failed to produce a valid "
                "toolchain",
                entry_point, entry_point.dist,
            )
            return

        if spec.get(EXPORT_TARGET) != export_target:
            logger.error(
                "the builder referenced by the entry point '%s' "
                "from package '%s' failed to produce a spec with the "
                "expected export_target",
                entry_point, entry_point.dist,
            )
            return

        if callable(verifier):
            warnings.warn(
                "%s:%s.verify_export_target returned a callable, which "
                "will no longer be passed to spec.advise by calmjs-4.0.0; "
                "please instead override 'setup_export_location' or "
                "'prepare_export_location' in that class" % (
                    self.__class__.__module__, self.__class__.__name__),
                DeprecationWarning
            )
            spec.advise(BEFORE_PREPARE, verifier, export_target)
        else:
            spec.advise(
                BEFORE_PREPARE,
                self.prepare_export_location, export_target)
        yield entry_point, toolchain, spec

    def iter_builders_for(self, package_name):
        for entry_point, export_target in self.iter_export_targets_for(
                package_name):
            # yield from...
            for builder in self.generate_builder(entry_point, export_target):
                yield builder

    def execute_builder(self, entry_point, toolchain, spec):
        """
        Accepts the arguments provided by the builder and executes them.
        """

        toolchain(spec)
        if not exists(spec['export_target']):
            logger.error(
                "the entry point '%s' from package '%s' failed to "
                "generate an artifact at '%s'",
                entry_point, entry_point.dist, spec['export_target']
            )
            return {}
        return self.generate_metadata_entry(entry_point, toolchain, spec)

    def process_package(self, package_name):
        results = {}
        for builder in self.iter_builders_for(package_name):
            results.update(self.execute_builder(*builder))
        return results


class ArtifactRegistry(BaseArtifactRegistry):
    """
    A registry to allow a central place for Python packages to declare
    which method to generate the artifact and what name to use.
    """

    def process_package(self, package_name):
        """
        Build artifacts declared for the given package.
        """

        metadata = super(ArtifactRegistry, self).process_package(package_name)
        if metadata:
            self.update_artifact_metadata(package_name, metadata)


class ArtifactBuilder(object):
    """
    A generic artifact builder.  Provides a callable object that will
    accept a list of package_names to generate prebuilt artifacts for
    based on the registry name used to construct this builder instance.
    """

    def __init__(self, registry_name=ARTIFACT_REGISTRY_NAME):
        """
        Arguments:

        registry_name
            the registry used to generate the artifacts for.
        """

        self.registry_name = registry_name

    def __call__(self, package_names):
        """
        Generic artifact builder function.

        Arguments:

        package_names
            List of package names to be built

        Returns True if the build is successful without errors, False if
        errors were found or if no artifacts were built.
        """

        result = True
        registry = get(self.registry_name)
        for package_name in package_names:
            metadata = {}
            for entry_point, export_target in registry.iter_export_targets_for(
                    package_name):
                builder = next(registry.generate_builder(
                    entry_point, export_target), None)
                if not builder:
                    # immediate failure if builder does not exist.
                    result = False
                    continue
                entries = registry.execute_builder(*builder)
                # whether the builder produced an artifact entry.
                result = bool(entries) and result
                metadata.update(entries)
            # whether the package as a whole produced artifacts entries.
            result = bool(metadata) and result
            registry.update_artifact_metadata(package_name, metadata)
        return result


class build_calmjs_artifacts(BuildArtifactCommand):
    """
    The main artifact build command for calmjs
    """

    artifact_builder = ArtifactBuilder(registry_name=ARTIFACT_REGISTRY_NAME)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from codecs import open
from os.path import join
from pkg_resources import resource_filename

from calmjs.parse.asttypes import String
from calmjs.parse.asttypes import Object
from calmjs.parse import es5
from calmjs import interrogate

# an example artifact bundle that concatenated both UMD and AMD together
artifact = """
(function () {
    (function umd(root, factory) {
        if(typeof exports === 'object' && typeof module === 'object')
            module.exports = factory();
        else if(typeof define === 'function' && define.amd)
            define('lib1', [], factory);
        else if(typeof exports === 'object')
            exports["lib1"] = factory();
        else
            root["lib1"] = factory();
    })(this, function() {});
    (function(define) {
        define('lib2', ['require', 'exports', 'module'], function (
            require, exports, module) {
        });
    }(
        typeof module === 'object' &&
        module.exports &&
        typeof define !== 'function' ? function (factory) {
            module.exports = factory(require, exports, module);
        } : define
    ));
}());
"""

# enclosed define.
artifact_multiple1 = """
(function (define) {(function(define) {
    define('lib1', ['require', 'exports', 'module', 'lib2'], function (
        require, exports, module
    ) {
        var lib2 = require('lib2');
    });
})(define);

define('text', ['module'], function (module) {
});

define('lib2', ['require', 'exports', 'module'], function () {
});

}(define));
"""

# standard AMD
artifact_multiple2 = """
define('lib1',[],function () {
    require('lib2');
    require('lib3');
});

define('lib2',[],function () {
    require('lib3');
});

define('lib4',[],function () {
    require('lib1');
});

define('lib3',[],function () {
});
"""

# missing case
artifact_multiple3 = """
define('lib1',[],function () {
    require('lib4');
    require('lib2');
    require('missing');
});

define('lib2',[],function () {
    require('lib4');
    require('missing');
});

define('lib4',[],function () {
    require('missing');
});
"""

# redefinition case
artifact_multiple4 = """
define('lib1',[],function () {
    require('lib3');
});

define('lib1',[],function () {
    require('missing');
});

define('lib3',[],function () {
});
"""

# commonjs
commonjs_require = """
var mod1 = require('mod1');
var mod2 = require("name/mod/mod2");
var invalid = require();

// test out dynamic require calls.
var target = mod1.target;
var dynamic = require(target);
"""

# require/define mix
requirejs_require = """
require(['some/dummy/module1', 'some/dummy/module2'], function(mod1, mod2) {
    var mod1_alt = require('some/dummy/module1');
    var mod3_alt = require('some/dummy/module3');
});

define(['defined/alternate/module', 'some.weird.module'], function(a, b) {
});

// this is a weird one
define('some test', ['require', 'module'], function(require, module) {});

// invalid code shouldn't choke the walker.
require();
define();

// test out dynamic define calls.
var target = window.target;
var dynamic = define([target], function(target_mod) {});
"""

# amd dynamic define
requirejs_dynamic_define = """
define(a_name, ['static1', 'static2'], function(static1) {
});
"""


class SimpleConversionTestCase(unittest.TestCase):
    """
    Test the various edge cases for the simple conversion.

    Note that the LHS (encapsulated in a String) is the raw data, while
    the RHS the quote style is similar to what JavaScript also supports,
    which is also similar to what Python does.
    """

    def test_to_str_basic(self):
        self.assertEqual(interrogate.to_str(String("'hello'")), 'hello')
        self.assertEqual(interrogate.to_str(String('"hello"')), 'hello')
        # Python escaped
        self.assertEqual(interrogate.to_str(String("'hell\"o'")), 'hell"o')
        self.assertEqual(interrogate.to_str(String('"hell\'o"')), "hell'o")

    def test_to_str_backslash(self):
        # JavaScript escaped
        self.assertEqual(interrogate.to_str(String(r"'he\'llo'")), 'he\'llo')
        self.assertEqual(interrogate.to_str(String(r"'he\"llo'")), 'he\"llo')
        self.assertEqual(interrogate.to_str(String(r"'he\\llo'")), 'he\\llo')

        self.assertEqual(interrogate.to_str(String(r'"he\'llo"')), "he\'llo")
        self.assertEqual(interrogate.to_str(String(r'"he\"llo"')), "he\"llo")

    def test_identifier_extract_typical(self):
        with open(resource_filename('calmjs.testing', join(
                'names', 'typical.js')), encoding='utf-8') as fd:
            tree = es5(fd.read())
        for obj_node in interrogate.shallow_filter(
                tree, lambda node: isinstance(node, Object)):
            self.assertEqual('typical', interrogate.to_identifier(
                obj_node.properties[0].left))

    def test_identifier_extract_unusual(self):
        answers = [
            '\'"\'',
            '\'\"\'',
            "\'\"\'",
            "\n",
            "\t",
            r'\\ ',
            "\u3042",
            "\u3042",
            "    ",
        ]
        with open(resource_filename('calmjs.testing', join(
                'names', 'unusual.js')), encoding='utf-8') as fd:
            tree = es5(fd.read())
        for obj_node, answer in zip(interrogate.shallow_filter(
                tree, lambda node: isinstance(node, Object)), answers):
            self.assertEqual(answer, interrogate.to_identifier(
                obj_node.properties[0].left))


class RequireJSHelperTestCase(unittest.TestCase):
    """
    Helpers to make this thing not suck like the other.
    """

    def test_extract_function_argument_basic(self):
        results = interrogate.extract_function_argument("""
        trial(1, 2, 'hello');
        trial(1, 2, "goodbye");
        """, 'trial', 2)
        self.assertEqual(results, ['hello', 'goodbye'])

    def test_extract_function_argument_mismatches(self):
        results = interrogate.extract_function_argument("""
        trial(1, 2);
        trial(1, 2, 23, 4, 5);
        """, 'trial', 2)
        self.assertEqual(results, [])

    def test_extract_function_argument_not_nested(self):
        # this is a very naive analysis; a more proper one will follow
        # through the scoping rules and trace through the variables so
        # that only the correctly bounded function calls be returned.
        results = interrogate.extract_function_argument("""
        (function() {
            trial(1, 2, 'hello', trial(1, 2, 'goodbye'));
            trial(1, 2, (function() { trial(1, 2, 'goodbye')})());
        })();
        """, 'trial', 2)
        self.assertEqual(results, ['hello'])

    def test_extract_function_not_sub(self):
        results = interrogate.extract_function_argument("""
        (function() {
            log('hello');
            log('');
            console.log('goodbye');
        })();
        """, 'log', 0)
        self.assertEqual(results, ['hello', ''])

    def test_extract_on_syntax_error(self):
        with self.assertRaises(SyntaxError):
            interrogate.extract_function_argument("""
            (function() {
                console.log('hello!');
                report('');
                missing_rparen(1, 2, 'hello';
            })();
            """, 'report', 0)

    def test_yield_imports_bad_type(self):
        with self.assertRaises(TypeError):
            next(interrogate.yield_module_imports('string = "invalid";'))

    def test_yield_imports_basic(self):
        self.assertEqual([
            'mod1',
            "name/mod/mod2",
        ], sorted(set(interrogate.yield_module_imports(es5(
            commonjs_require)))))

    def test_extract_all_commonjs_requires(self):
        self.assertEqual([
            'mod1',
            "name/mod/mod2",
        ], sorted(set(interrogate.extract_module_imports(commonjs_require))))

    def test_extract_all_amd_requires(self):
        self.assertEqual([
            'defined/alternate/module',
            'some.weird.module',
            'some/dummy/module1',
            'some/dummy/module2',
            'some/dummy/module3',
        ], sorted(set(interrogate.extract_module_imports(requirejs_require))))

    def test_extract_all_dynamic_amd_define_require(self):
        self.assertEqual([
            'static1',
            'static2',
        ], sorted(set(interrogate.extract_module_imports(
            requirejs_dynamic_define
        ))))

    def test_yield_imports_node_bad_type(self):
        with self.assertRaises(TypeError):
            next(interrogate.yield_module_imports_nodes('string = "invalid";'))

    def test_dynamic_amd_define_require_yield_node(self):
        self.assertEqual(2, len(list(interrogate.yield_module_imports_nodes(
            es5(requirejs_dynamic_define)))))

    def test_dynamic_define_dynamic_cjs_require(self):
        self.assertEqual(2, len(list(interrogate.yield_module_imports_nodes(
            es5("""
            var foobar = 'module';
            var mod = require(foobar);
            var mod2 = require('foobar');
            var invalid = require(foobar, 'foobar');
            """)
        ))))

    def test_dynamic_define_dynamic_amd_require(self):
        self.assertEqual(2, len(list(interrogate.yield_module_imports_nodes(
            es5("""
            require([foobar, 'foobar'], function(dyn, static) {
            });
            """)
        ))))

    def test_extract_all_amd_requires_skip_reserved(self):
        src = (
            "define('some/test', ['require', 'exports', 'module'], "
            "function(require, exports, module) {});"
        )
        self.assertEqual(
            [],
            sorted(set(interrogate.extract_module_imports(src)))
        )

        src = (
            "define('some/test', ['require', 'exports', 'module', 'mod1'], "
            "function(require, exports, module, mod1) {});"
        )
        self.assertEqual(
            ['mod1'],
            sorted(set(interrogate.extract_module_imports(src)))
        )

        # an actual real world one with the loader plugin
        src = "define('text', ['module'], function(module) {});"
        self.assertEqual(
            [],
            sorted(set(interrogate.extract_module_imports(src)))
        )

        src = "define('alt/text', ['module', 'alt'], function(module) {});"
        self.assertEqual(
            ['alt'],
            sorted(set(interrogate.extract_module_imports(src)))
        )

        src = "define('alt/text', ['alt', 'module'], function(a, module) {});"
        self.assertEqual(
            ['alt'],
            sorted(set(interrogate.extract_module_imports(src)))
        )

    def test_extract_all_amd_requires_non_standard(self):
        # shouldn't fail on certain non-standard cases.
        src = (
            "define('some/test', [], function(require, exports, module) {});"
        )
        self.assertEqual(
            [],
            sorted(set(interrogate.extract_module_imports(src)))
        )

        src = (
            "require();"
        )
        self.assertEqual(
            [],
            sorted(set(interrogate.extract_module_imports(src)))
        )



# -*- coding: utf-8 -*-
"""
Module for providing various distutil command that integrates with
setuptools, and for setting up the build environments and building
JavaScript artifacts.
"""

from __future__ import absolute_import

import sys
import logging

from distutils.errors import DistutilsModuleError
from distutils.core import Command
from distutils import log

from calmjs.ui import prompt_overwrite_json


class DistutilsLogHandler(logging.Handler):
    """
    A handler that streams the logs to the distutils logger.
    """

    def __init__(self, distutils_log=log):
        logging.Handler.__init__(self)
        self.log = distutils_log
        # Basic numeric table
        self.level_table = {
            logging.CRITICAL: distutils_log.FATAL,
            logging.ERROR: distutils_log.ERROR,
            logging.WARNING: distutils_log.WARN,
            logging.INFO: distutils_log.INFO,
            logging.DEBUG: distutils_log.DEBUG,
        }
        self.setFormatter(logging.Formatter('%(message)s'))

    def _to_distutils_level(self, level):
        return self.level_table.get(level, level // 10)

    def emit(self, record):
        level = self._to_distutils_level(record.levelno)
        try:
            msg = self.format(record)
            self.log.log(level, msg)
        except Exception:  # pragma: no cover
            # LogRecord.__str__ shouldn't fail... probably.
            self.log.warn('Failed to convert %s' % record)


distutils_log_handler = DistutilsLogHandler()


def use_distutils_logger(logger_ids=('calmjs',)):
    def decorator(method):
        def run(cmd):
            root_logger = logging.getLogger()
            old_level = root_logger.level
            root_logger.setLevel(logging.DEBUG)

            for logger_id in logger_ids:
                logger = logging.getLogger(logger_id)
                logger.addHandler(distutils_log_handler)

            try:
                method(cmd)
            finally:
                # Remove the logging handlers and restore the level.
                for logger_id in logger_ids:
                    logger = logging.getLogger(logger_id)
                    logger.removeHandler(distutils_log_handler)

                root_logger.setLevel(old_level)

        return run
    return decorator


class PackageManagerCommand(Command):
    """
    Simple compatibility hook for a package manager runtime.
    """

    # subclasses need to define these
    runtime = None
    # description = "base command for package manager compatibility helper"

    indent = 4

    @classmethod
    def _initialize_user_options(cls):
        cls.user_options = []
        for full, short, desc in cls.runtime.pkg_manager_options:
            if short is None:
                cls.user_options.append((full, short, 'action: ' + desc))
            else:
                cls.user_options.append((full, short, desc))

    # keywords that are actions that result in effects that we support
    # TODO derive this like how the runtime does.
    actions = ('view', 'init', 'install')

    def _opt_keys(self):
        for opt in self.user_options:
            yield opt[0]

    def initialize_options(self):
        for key in self._opt_keys():
            setattr(self, key, False)
        self.stream = None  # extra output
        self.callback = None

    def do_view(self):
        pkg_name = self.distribution.get_name()
        self.cli_driver.pkg_manager_view(pkg_name, stream=self.stream)

    def do_init(self):
        pkg_name = self.distribution.get_name()
        self.cli_driver.pkg_manager_init(
            pkg_name,
            overwrite=self.overwrite, merge=self.merge,
            callback=self.callback,
            stream=self.stream,
        )

    def do_install(self):
        pkg_name = self.distribution.get_name()
        self.cli_driver.pkg_manager_install(
            pkg_name,
            overwrite=self.overwrite, merge=self.merge,
            callback=self.callback,
            production=self.production, development=self.development,
            stream=self.stream,
        )

    def finalize_options(self):
        opts = [i for i in (getattr(self, k) for k in self.actions) if i]
        if not opts:
            # default to view
            self.view = True
        if self.view or self.dry_run:
            self.stream = sys.stdout
        self.callback = prompt_overwrite_json if self.interactive else None
        # require explicit boolean value.
        self.production = True if self.production else None
        self.development = True if self.development else None

    @use_distutils_logger()
    def run(self):
        if self.dry_run:
            # Do the default action and finish, as everything else may
            # cause permanent changes.
            self.do_view()
            return
        self.run_command('egg_info')
        if self.install:
            self.do_install()
        elif self.init:
            self.do_init()
        elif self.view:
            self.do_view()


class BuildArtifactCommand(Command):
    """
    Command for building artifacts for the given package.
    """

    user_options = []
    artifact_builder = None

    def initialize_options(self):
        """
        Implement items that could go into the spec, such as setting of
        build dir prefix.
        """

    def finalize_options(self):
        """
        If finalization is needed.
        """

    @use_distutils_logger()
    def run(self):
        if not callable(self.artifact_builder):
            raise DistutilsModuleError(
                "artifact_builder is not callable for %s" % self)
        if self.dry_run:
            return
        package_name = self.distribution.get_name()
        if not self.artifact_builder([package_name]):
            raise DistutilsModuleError(
                "some entries in registry '%s' defined for package '%s' "
                "failed to generate an artifact" % (
                    self.artifact_builder.registry_name,
                    package_name,
                )
            )
